/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"
/* epoll-specific parts see queue_common.c and activeq.c for the rest */
/*
 * a poll/select/libev/libevent-based implementation would have a hard time
 * migrating clients between threads
 */
#ifdef HAVE_EPOLL_WAIT
struct mog_queue * mog_queue_new(void)
{
	int size_hint = 666; /* hint, ignored in new kernels */
	int epoll_fd = epoll_create(size_hint);
	if (epoll_fd < 0) die("epoll_create failed: %s\n", strerror(errno));

	return mog_queue_init(epoll_fd);
}

/*
 * grabs one active event off the event queue
 * epoll_wait() has "wake-one" behavior (like accept())
 * to avoid thundering herd since 2007
 */
struct mog_fd * mog_idleq_wait(struct mog_queue *q, int timeout)
{
	int rc;
	struct epoll_event event;
	struct mog_fd *mfd;

retry:
	/* epoll_wait is a cancellation point since glibc 2.4 */
	rc = epoll_wait(q->queue_fd, &event, 1, timeout);
	switch (rc) {
	case 1:
		mfd = event.data.ptr;
		mog_fd_check_out(mfd);
		return mfd;
	case 0:
		return NULL;
	}

	switch (errno) {
	case EINTR:
		goto retry;
	case EBADF:
	case EINVAL: /* epfd can be hit */
		return NULL;
	}

	die("epoll_wait failed with (%d): %s\n", rc, strerror(errno));
	return NULL;
}

MOG_NOINLINE static void
epoll_ctl_error(struct mog_queue *q, struct mog_fd *mfd)
{
	switch (errno) {
	case EPERM:
	case EBADF:
		/* TODO: check for shutdown races */
		syslog(LOG_ERR, "bad file descriptor for epoll_ctl()");
		return;
	case ENOSPC:
		syslog(LOG_ERR,
		       "epoll out-of-space, falling back to active queue");
		mog_activeq_push(q, mfd);
		return;
	case ENOMEM: mog_oom();
	case ENOENT:
	case EEXIST:
	case EINVAL:
		syslog(LOG_ERR, "unhandled epoll_ctl() error: %m");
		assert(0 && "BUG in our usage of epoll");
	default:
		syslog(LOG_ERR, "unhandled epoll_ctl() error: %m");
	}
}

/*
 * Pushes in one mog_fd for epoll to watch.
 *
 * Only call this from the mog_accept_loop *or*
 * if EAGAIN/EWOULDBLOCK is encountered in mog_queue_loop.
 */
void mog_idleq_push(struct mog_queue *q, struct mog_fd *mfd, enum mog_qev ev)
{
	struct epoll_event event;
	int op = mfd->queue_state == MOG_QUEUE_STATE_OLD ?
	         EPOLL_CTL_MOD : EPOLL_CTL_ADD;

	mfd->queue_state = MOG_QUEUE_STATE_OLD;
	event.data.ptr = mfd;
	event.events = (int)ev;

	mog_fd_check_in(mfd);
	if (epoll_ctl(q->queue_fd, op, mfd->fd, &event) != 0) {
		mog_fd_check_out(mfd);
		epoll_ctl_error(q, mfd);
	}
}
#else /* ! HAVE_EPOLL_WAIT */
typedef int avoid_empty_file;
#endif /* ! HAVE_EPOLL_WAIT */
