/*
 * Copyright (C) 2012, Eric Wong <normalperson@yhbt.net>
 * License: GPLv3 or later (see COPYING for details)
 */
#include "cmogstored.h"

struct qcleanup {
	struct mog_queue *queue;
	struct mog_fd **active_mfd;
};

static void queue_loop_cleanup(void *arg)
{
	struct qcleanup *cleanup = arg;
	struct mog_queue *q = cleanup->queue;
	struct mog_fd *mfd = *cleanup->active_mfd;
	unsigned long self = (unsigned long)pthread_self();

	syslog(LOG_DEBUG, "mog_queue_loop[%lx] thread shutting down...", self);

	/* step through the client we didn't want to share, first */
	while (mfd)
		mfd = mog_queue_step(mfd);

	/*
	 * in case an accept loop thread pushed into idle queue right
	 * before shutdown, we need to account for clients here
	 */
	while ((mfd = mog_idleq_wait(q, 0))) {
		do {
			mfd = mog_queue_step(mfd);
		} while (mfd);
	}
	/*
	 * we need to drain the active queue to prevent starvation
	 * if we are reducing worker threads (and this thread is
	 * getting cut).  The reason we normally avoid deadlocks
	 * in other cases is that the thread that pushed into
	 * active queue is always capable of taking the same mfd
	 * from the active queue later.
	 */
	while ((mfd = mog_activeq_trytake(q))) {
		do {
			mfd = mog_queue_step(mfd);
		} while (mfd);
	}

	syslog(LOG_DEBUG, "mog_queue_loop[%lx] thread done", self);
}

static struct mog_fd *cancellable_queue_wait(struct mog_queue *q, int timeout)
{
	struct mog_fd *mfd;

	mog_cancel_enable();
	mfd = mog_idleq_wait(q, timeout);
	mog_cancel_disable();

	return mfd;
}

/* passed as a start_routine to pthread_create */
void * mog_queue_loop(void *arg)
{
	struct mog_queue *q = arg;
	struct mog_fd *active_mfd = NULL;
	struct qcleanup cleanup;

	cleanup.queue = q;
	cleanup.active_mfd = &active_mfd;

	pthread_cleanup_push(queue_loop_cleanup, &cleanup);
	mog_cancel_disable();
	syslog(LOG_DEBUG, "mog_queue_loop[%lx] thread ready",
	       (unsigned long)pthread_self());

	for (;;) {
		struct mog_fd *mfd;

		/*
		 * idle, just-ready clients are the most important
		 * handle them as much as possible and bounce them
		 * into the active queue...
		 */
		while ((mfd = cancellable_queue_wait(q, 0))) {

			/*
			 * We got a more important client, push active_mfd
			 * into the active queue for another thread to while
			 * we service another client.
			 */
			if (active_mfd)
				mog_activeq_push(q, active_mfd);
			active_mfd = mog_queue_step(mfd);
		}

		/*
		 * Try to avoid locking the global active queue and keep
		 * working on the same active_mfd for as long as possible.
		 */
		if (active_mfd) {
			active_mfd = mog_queue_step(active_mfd);

		/* see if there's any already-active clients to work on: */
		} else if ((mfd = mog_activeq_trytake(q))) {
			active_mfd = mog_queue_step(mfd);
		} else {
			/*
			 * We'll get here if there's nothing to do.
			 * Sleep until there's an event.  mog_accept_loop
			 * will push into epoll/kqueue to wake us up here.
			 */
			if ((mfd = cancellable_queue_wait(q, -1)))
				active_mfd = mog_queue_step(mfd);
		}
	}

	pthread_cleanup_pop(1);

	return NULL;
}

static void queue_quit_step(struct mog_fd *mfd)
{
	switch (mfd->fd_type) {
	case MOG_FD_TYPE_MGMT: mog_mgmt_quit_step(mfd); return;
	case MOG_FD_TYPE_HTTP:
	case MOG_FD_TYPE_HTTPGET:
		mog_http_quit_step(mfd); return;
	case MOG_FD_TYPE_FILE:
	case MOG_FD_TYPE_QUEUE:
	case MOG_FD_TYPE_SVC:
		assert(0 && "invalid fd_type in queue_quit_step");
	default:
		break;
	}
}

/* called at shutdown when only one thread is active */
void mog_queue_quit_loop(struct mog_queue *queue)
{
	struct mog_fd *mfd;

	while (mog_nr_active_at_quit) {
		assert(mog_nr_active_at_quit <= (size_t)INT_MAX
		       && "mog_nr_active_at_quit underflow");

		if ((mfd = mog_activeq_trytake(queue))) {
			queue_quit_step(mfd);
		} else {
			if ((mfd = mog_idleq_wait(queue, -1)))
				queue_quit_step(mfd);
		}
	}
}
